#Asm class: Assemble a listing from a .asm file.
import sys
import os
from Hc11set import *

class AsmError:
    def __init__(self, errtype, lnum):
        self.errtype = errtype
        self.lnum = lnum

OffsetError = 'Offset out of range.'
FRTypeError = 'Forward reference error.'
AsmErrorModes = set(['Immediate','Direct','Extended','Index,X','Index,Y','Inherent','Relative'])

class Asm:
    """Asm()
    Any instance of this class use this classes attributes as 'global' variables,
    changing them freely. The unique thing about an instance is its asmfile.  If
    this file requires assembly of another file, simply create a new instance or
    Asm and pass in the file name and the current 'listing'."""

    ## Initialize class attributes which act as 'globals' for each instance of Asm
    constants = {}  # list of labels turned into Python variables for EVALuation
    labels = {}     # dictionary of resolved labels
    forwards = {}   # dictionary of forward references:{[(linenum,addrmode,pcounter)
    pcounter = 0    # Program Counter
    linenum = 0     # Index into listing. Equal to Line Number - 1
    includes = []

    def __init__(self, listing):
        Asm.listing = listing
        # Instance variables
        self.errors = 0      # Error counter or flag; cleared here globally
        self.curlabel = ''
        self.printlist = False    # Default NOT to print the listing as it's assembled
        self.comments = False    # Default NOT to print comments

    def AsmLine(self, line):
    # Functions
        def Rem1Word(string):       # Remove 1 word from the begining of string;
            if not string or type(string) != str:
                return '', ''
            s=''
            for i, a in enumerate(string):
                if a.isspace(): break
                s += a
            return s, string[i + 1:].lstrip(' \t')

        def GetMoreOperands(a, mne, operands, mode):
            if not ((mne in TwoOperands) or (mne in ThreeOperands)):
                raise AsmError('Too many operands for %s addressing' %mode, Asm.linenum)
            mask = GetCOrEVal(operands[0])
            if mask == None:
                raise AsmError('Bad second operand for %s addressing' %mode, Asm.linenum)
            Asm.pcounter += 1
            hexop = hex(mask)[2:].upper().zfill(2)
            if mne in ThreeOperands:
                if a < 2:
                    raise AsmError('Too few operands for %s addressing' %mode, Asm.linenum)
                operand1 = operands[1]
                Asm.pcounter += 1
                if operand1 == '*':
                    dest = hex(oldPC + 256 - Asm.pcounter)[2:].upper().zfill(2)
                else:
                    dest = Offset(operand1)
                    if dest == None:
                        raise AsmError('Bad third operand for %s addressing' %mode, Asm.linenum)
                hexop += dest
            return hexop

        def Invert(hexorint):
            "Invert a hex number (or an integer < 65536), return an int"
            if type(hexorint) == str:
                a = len(hexorint) - 2
                return ~int(hexorint,16)+16**a
            elif type(hexorint) == int:
                a = 1 + (hexorint > 15)
                a += (hexorint > 255)
                a += (hexorint > 4095)
                return ~hexorint + 16**a

        def MakeNum(numstr):
            "Convert Assembler number strings to Python numbers"
            if numstr.isdigit(): return int(numstr)
            if len(numstr) < 2: return None
            if numstr[0] == '$': return '0x' + numstr[1:]   # hex
            if numstr[0] == '@': return int(numstr[1:],8)   # octal
            if numstr[0] == '%': return int(numstr[1:],2)   # binary

        def GetHexByteOrWord(value):
            "Return hex digit string with length based on magnitude of int value"
            a = Byte + (value > 255)
            return hex(value)[2:].upper().zfill(2 * a)

        def GetCOrEVal(operands):    # Is the operand field an expression or a constant?
            "Return value of an expression/constant or None if operands can't be evaluated."
            if operands[0] in Quotes:
                return ord(operands[1])   # ascii for A or B
            for oper in Operators.intersection(operands):
                l = operands.split(oper)
                operands = (' %s ' %oper).join(l)
            expression = operands.split()
            for i, expr in enumerate(expression):   # Pull out asm-style numbers
                invert = False
                if expr in Operators or expr.isdigit():
                    continue
                if expr[0] == '!':    # One's compliment
                    invert = True
                    expr = expr[1:]
                num = MakeNum(expr)       # If it is not an operator or a numstring
                if num:
                    a = str(num)   # Replace text with a number string
                else:
                    try:
                        a = str(eval('Asm.%s' %expr))
                    except (NameError, AttributeError, SyntaxError):
                        return None
                if invert:
                    a = str(Invert(int(a)))
                expression[i] = a
            try:
                a = eval(' '.join(expression))   # The line can be evaluated
                if a < 0:
                    a = Invert(int(abs(a) - 1))
            except (NameError, AttributeError, SyntaxError):
                return None
            return a

        def CheckOp(opfield, size):     # Check evaluability and magnitude of opfield
            """Raise an AsmError if the opfield can't be evaluated or is too big.
            Return the result otherwise."""
            a = GetCOrEVal(opfield)
            if a == None:
                raise AsmError('Bad operand', Asm.linenum)
            if a > (256**size) -1:
                raise AsmError('Operand too big for %d bits' %(8 * size), Asm.linenum)
            return a

        def CheckHex(opfield, size):    # Check evaluability and magnitude of opfield
            """Raise an AsmError if the opfield can't be evaluated or is too big.
            Return the hex equiv. with '0x' stripped off otherwise."""
            return hex(CheckOp(opfield, size))[2:].upper().zfill(size * 2)

        def MakeForward(label, addmod, ftype):   # Install forward reference
            """Add or append a new forward reference to this labels list.
            Return placeholder characters"""
            if label in Asm.forwards:
                Asm.forwards[label].append((Asm.linenum, addmod, ftype))
            else:
                Asm.forwards[label] = [(Asm.linenum, addmod, ftype)]
            return '??'

        def GetEffAddr(label, addmod, mne, size): # Direct and Extended mode addresses
            """Return the Effective Address of a label in hex or sized placeholder
            for forward references"""
            ea = Asm.labels.get(label)  # If the label exists
            if ea:
                if ea > (256**size) - 1:    # If its value isn't too big
                    return None
                return hex(ea)[2:].upper().zfill(size * 2)   # Return the hex representation
            return MakeForward(label, addmod, mne) * size

        def Offset(label):      # Relative address mode offset from PC
            """Return the 8 bit backward offset to a label in hex, or a forward reference
            placeholder"""
            if label in Asm.constants:
                return None
            i = Asm.labels.get(label)   # If the label exists
            if i:                   # Calculate BACKWARD references #
                j = i - Asm.pcounter    # i is always a lower address than PCounter
                if j < -128:
                    raise AsmError(OffsetError, Asm.linenum)
                return hex(i + 256 - Asm.pcounter)[2:].upper().zfill(2)
            return MakeForward(label, Relative, Asm.pcounter)

        def MakeLabel(label):    # Left side "label"
            """When a label definition occurrs, install in labels dictionary and
            resolve any forward references"""
            if label in Asm.labels or label in Asm.constants:
                raise AsmError('Redefined label', Asm.linenum)
            Asm.labels[label] = Asm.pcounter
            if label in Asm.forwards:   # Resolve FORWARD references #
                for i in range(len(Asm.forwards[label])):
                    fline, mode, addr = Asm.forwards[label].pop()
                    if mode in Branches:
                        value = Asm.pcounter - addr # PCounter is always a higher address than addr
                        if value > 127 :
                            raise AsmError(OffsetError, Asm.linenum)
                        a = Byte
                    else:
                        Asm.labels[label] = Asm.pcounter
                        value = Asm.pcounter
                        a = Word
                    Asm.listing.FixLine(fline, a, value)
                Asm.forwards.pop(label)

        def MakeConst(label, value):    # Left side "label"
            """When a constant definition occurrs, install in constants list and
            resolve any forward references"""
            if label in Asm.constants or label in Asm.labels:
                raise AsmError('Redefined constant', Asm.linenum)
            Asm.constants[label] = value
            if label in Asm.forwards:   # Resolve FORWARD references #
                for i in range(len(Asm.forwards[label])):
                    fline, mode, mne = Asm.forwards[label].pop()
                    a = Byte + ((mode == Extended) or (value > 255) or (mne in TwoBytes))
                    Asm.listing.FixLine(fline, a, value)
                Asm.forwards.pop(label)
            try:
                exec(label + '=' + str(value)) in Asm.__dict__
            except (NameError, SyntaxError):
                raise AsmError('Could not resolve %s = %d on line' %(label, value), Asm.linenum)

        ### AsmLine() ###
        # Parse out the label, mnemonic and operand
        label, line = Rem1Word(line)
        mne, line = Rem1Word(line)   # opfield + comments + \n left in line
        if not mne:
            raise AsmError('Invalid empty mnemonic field', Asm.linenum)
        mne = mne.lower()
        a = line.find('\t')
        if a < 0:
            a = line.find(';')
        opfield = line[:a].rstrip()  # Opfield has only operand(s)
        line = line[a:-1].lstrip()

        oldPC = Asm.pcounter
        addrmode = -1

    ###  Process Psudo-Op mnemonics  ##
        if mne in Psudos:

            if mne == 'equ':        # EQUate
                if not (label or opfield):
                    raise AsmError('Syntax error for EQUate', Asm.linenum)
                if opfield[0] == '*':
                    operand1 = Asm.pcounter
                    hexcode = hex(operand1)[2:].upper().zfill(4)
                    MakeLabel(label)
                else:
                    operand1 = CheckOp(opfield, Word)
                    hexcode = hex(operand1)[2:].upper().zfill(4)
                    MakeConst(label, operand1)
                return (oldPC, label, '', hexcode, mne, opfield, line)

            if mne == 'fcc':      # Form Constant Characters (strings)
                if not opfield or not (opfield[0] in Quotes and opfield[-1] in Quotes):
                    raise AsmError('Operand is not a string', Asm.linenum)
                templab = label
                tempfld = opfield
                if label:
                    pc = Asm.pcounter
                    MakeConst(label, Asm.pcounter)
                opfield = opfield[1:-1]
                a = len(opfield)
                for i in range(0, a, 8):
                    opcode = ''.join(map((lambda char: hex(ord(char))
                                 [2:].upper().zfill(2)), opfield[i:i + 4]))
                    hexcode = ''.join(map((lambda char: hex(ord(char))
                                 [2:].upper().zfill(2)), opfield[i + 4:i + 8]))
                    Asm.linenum += 1
                    Asm.listing.AddLine(Asm.linenum, (oldPC, templab, opcode,
                                        hexcode, mne, tempfld, line), self.printlist)
                    templab = ' .  ' * 3
                    mne = '.   .'
                    tempfld, line = '', ''
                    Asm.pcounter += (len(opcode)/2) + (len(hexcode)/2)
                    oldPC = Asm.pcounter
                return None

            if mne in ['fcb', 'fdb']:    # Form Constant Byte or Form Double Byte
                size = 1 + (mne == 'fdb')
                if label:
                    pc = Asm.pcounter
                    MakeConst(label, pc)
                operands = opfield.split(',')
                if len(operands) > 8/size:
                    raise AsmError('Too many aruments to Form Constant Bytes', Asm.linenum)
                opcode = ''
                for oper in operands:
                    operand1 = GetCOrEVal(oper)  # Any number or expression
                    if operand1 == None:        # Forward reference to a constant or label for Jumps
                        opcode += GetEffAddr(oper, addrmode, mne, size)
                    else:
                        opcode += hex(operand1)[2:].upper().zfill(2 * size)
                    Asm.pcounter += size
 #               if len(opcode) > 16:
                hexop = opcode[8:]
                opcode = opcode[:8]
                return (oldPC, label, opcode, hexop, mne, opfield, line)

            if mne == 'rmb':        # Reserve Memory Byte
                operand1 = CheckOp(opfield, Byte)
                if label:
                    pc = Asm.pcounter
                    MakeConst(label, pc)
                hexcode = hex(Asm.pcounter)[2:].upper().zfill(4)
                Asm.pcounter += operand1
                return (oldPC, label, '', hexcode, mne, opfield, line)

            if mne == 'org':        # ORiGin
                operand1 = GetCOrEVal(opfield)
                if operand1 == None:
                    operand1 = GetEffAddr(opfield, -1, mne, Word)
                    if operand1 == None:
                        raise AsmError('Can\'t set ORiGin to %s' %opfield, Asm.linenum)
                    operand1 = int(operand1, 16)
                Asm.pcounter = operand1
                hexcode = hex(Asm.pcounter)[2:].upper().zfill(4)
                if label:
                    pc = Asm.pcounter
                    MakeConst(label, pc)
                return (oldPC, label, '', hexcode, mne, opfield, line)

            if mne == 'include':    ## Handle "include"d files here ##
                Asm.linenum += 1
                Asm.listing.AddLine(Asm.linenum, ('', label, '', '', mne,
                                    opfield, line), self.printlist)

                # if opfield has directroy info, use that
                # else get current working dirctory

                subasmfilename = '%s\\%s' %(os.getcwd(), opfield)
                print 'Including %s' % subasmfilename
                if not os.path.isfile(subasmfilename):
                    raise AsmError('%d couldn\'t be found' %subasmfilename, Asm.linenum)

                if not subasmfilename in Asm.includes:      # Allow modules to be assembled
                                                            # separately or en masse by keeping
                    Asm.includes.append(subasmfilename)     # track of what's been included
                    subasm = Asm(Asm.listing)       # Create a new instance of Asm
                    self.errors = subasm.AsmFile(subasmfilename)    # Run it
                    del subasm                                      # Delete it
                    if self.errors:
                        raise AsmError('Error in INCLUDEd file', Asm.linenum)
                return None

            if mne == 'option':
                if opfield == 'list_on':
                    self.printlist = True
                elif opfield == 'list_off':
                    self.printlist = False
                elif opfield == 'comments_on':
                    self.comments = True
                elif opfield == 'comments_off':
                    self.comments = False
                else:
                    raise AsmError('Unknown option specification: %s' %opfield, Asm.linenum)
                return (oldPC, label, '', '', mne, opfield, line)

    ####  Process label, HC11 mnemonics and operand field  ###
        if label:               # Labels MUST be created while PCounter still points
            if label[0] != ':':
                self.curlabel = label    # Automatic labels are created here
            else:
                label = self.curlabel[:12-len(label)] + label
            MakeLabel(label)    # to the code that is being assembled
        if opfield:
            if opfield[0] == ':':
                opfield = self.curlabel[:12-len(opfield)] + opfield

    # Inherent mode handled here
        if mne in InhCodes:
            addrmode = Inherent
            opcode = InhCodes[mne]
            Asm.pcounter += len(opcode)/2
            hexop = '' # Never an operand
            return (oldPC, label, opcode, hexop, mne, opfield, line)

        if not opfield:
            raise AsmError('Invalid empty operand field', Asm.linenum)

    # Relative mode handled here
        if mne in RelCodes:
            operand1, opfield = Rem1Word(opfield)
            if opfield or not operand1:
                raise AsmError('Bad label for relative address mode', Asm.linenum)
            addrmode = Relative
            opcode = RelCodes[mne]
            Asm.pcounter += 2   # 1 opcode + 1 byte operand added BEFORE offset calculation
            if operand1 == '*':
                hexop = hex(oldPC + 256 - Asm.pcounter)[2:].upper().zfill(2)
            else:
                if GetCOrEVal(operand1) != None:
                    raise AsmError('Illegal use of constant', Asm.linenum)
                hexop = Offset(operand1)
                if hexop == None:
                    raise AsmError('Bad operand for relative address mode', Asm.linenum)
            return (oldPC, label, opcode, hexop, mne, opfield, line)

        if not mne in MixModes:
            raise AsmError('Unrecognized mnemonic', Asm.linenum)

    # Immediate mode handled here
        if opfield[0] == '#':
            addrmode = Immediate
            a = Byte + (mne in TwoBytes)
            operand1 = GetCOrEVal(opfield[1:])  # Any number or expression
            if operand1 == None:        # Forward reference to a constant or label for Jumps
                hexop = GetEffAddr(opfield[1:], addrmode, mne, a)
                if not hexop:
                    raise AsmError('Bad operand for immediate address', Asm.linenum)
            else:
                if operand1 > 256**a - 1:
                    raise AsmError('Operand too big for %d bit address' %(8 * a), Asm.linenum)
                hexop = hex(operand1)[2:].upper().zfill(2 * a)
            Asm.pcounter += a
            opcode = MixModes[mne][addrmode]
            if opcode == 'NoOp':
                raise AsmError('Can\'t use %s address mode for this instruction' %mode, Asm.linenum)
            Asm.pcounter += len(opcode)/2
            return (oldPC, label, opcode, hexop, mne, opfield, line)

    # Indexed modes handled here
        if ',X' in opfield.upper():
            addrmode = IndexX
            a = opfield.upper().find(',X')
        elif ',Y' in opfield.upper():
            addrmode = IndexY
            a = opfield.upper().find(',Y')
        if addrmode == IndexX or addrmode == IndexY:    # Common code for both indexed address modes
            operand1 = opfield[:a]
            opfield = opfield[a + 2:].strip()   # Strip off the ,X i.e. and spaces
            Asm.pcounter += 1
            if not operand1:
                hexop = '00'    # Default value for ,Y i.e.
            else:
                operand1 = GetCOrEVal(operand1)    # Do the math
                if operand1 == None:
                    raise AsmError('Use constant for offset to index register', Asm.linenum)
                hexop = hex(operand1)[2:].upper().zfill(2)
            opcode = MixModes[mne][addrmode]
            if opcode == 'NoOp':
                raise AsmError('Can\'t use Indexed address mode for this instruction', Asm.linenum)
            Asm.pcounter += len(opcode)/2   # Increment PC BEFORE possible relative branch
            if opfield:     # Are there more operands to process?
                operands = opfield.split()
                a = len(operands) + 1   # The first operand was removed above
                if operands[0]:     # If there are more operands, go get 'em
                    hexop += GetMoreOperands(a, mne, operands, 'Direct')
            return (oldPC, label, opcode, hexop, mne, opfield, line)

    # Direct and Extended address modes handled here:
        hexop = addrmode = False    # Mode MAY be forced; hexop MAY come from a label
        am = Extended   # Default address mode for forward references
        operands = opfield.split()
        if not operands[0]:
            raise AsmError('Too few operands for direct or extended addressing', Asm.linenum)
        if operands[0][0] == '<':          # Set address mode if it is being forced
            operands[0] = operands[0][1:]  # And strip the force character
            addrmode = am = Direct      # Forward references MAY be forced to direct
        elif operands[0][0] == '>':
            operands[0] = operands[0][1:]
            addrmode = Extended
        operand1 = GetCOrEVal(operands[0])
        if operand1 == None:        # Forward reference to a constant or label
            hexop = GetEffAddr(operands[0], am, mne, 2)  # May have leading zeros
            if not hexop:
                raise AsmError('Bad operand for %s address'
                           %AsmErrorModes[addrmode], Asm.linenum)
            if hexop[0] == '?': # Is it a forward reference?
                if addrmode == Direct:
                    hexop = hexop[:2]
                addrmode = am
            else:   # No, not a forward reference
                operand1 = int(hexop,16)
        if not addrmode:
            addrmode = Direct + ((operand1 > 255) or (mne in ExtOnly))
        a = len(operands)
        if addrmode == Extended:  # Extended takes only one operand
            if a > 1:
                raise AsmError('Too many operands for extended addressing', Asm.linenum)
        if operand1 > 256**addrmode - 1:
            raise AsmError('Operand too big for %d bit address' %(8 * addrmode), Asm.linenum)
        if not hexop:
            hexop = hex(operand1)[2:].upper().zfill(addrmode * 2)
        Asm.pcounter += addrmode
        opcode = MixModes[mne][addrmode]
        if opcode == 'NoOp':
            raise AsmError('Can\'t use %s address mode for this instruction'
                           %AsmErrorModes[addrmode], Asm.linenum)
        Asm.pcounter += len(opcode)/2
        if a > 1:
            operands.pop(0)  # Make the list look like Indexed
            hexop += GetMoreOperands(a, mne, operands, 'Direct')
        return (oldPC, label, opcode, hexop, mne, opfield, line)

    def AsmFile(self, asmfilename):
        """On each line of the assembly file, copy comments and blank lines into listing
        VARIFY EACH LINE BETTER???????? then do assembler directives or call AsmLine()"""
        try:
            asmfile = open(asmfilename)
            for i, line in enumerate(asmfile):
                # Skip blank lines or comment lines
                # listline example: 10007 FFFF HHHHHHHH HHHHHHHH label(12) xgxy op, op, op ;comment
                if line.isspace() or line.lstrip()[0] in ['*', ';']:
                    if self.comments:
                        Asm.linenum += 1
                        Asm.listing.AddLine(Asm.linenum, ('', '', '', '',
                                            '', '', line[:-1]), self.printlist)
                    continue

                try:     # Call AsmLine() oldPC, label, opcode, hexop, mne, opfield, line
                    parts = self.AsmLine(line)
                except AsmError, instagator:    # Process all errors and break out of loop
                    self.errors = 1
                    print '\n',instagator.errtype
                    print 'File \"%s\", line %d' %(asmfile.name, i + 1)
                    print repr(line)
                    break

                if parts is not None:
                    Asm.linenum += 1    # Increment global line number
                    Asm.listing.AddLine(Asm.linenum, parts, self.printlist) # Add a line to listing

        finally:
            asmfile.close()
        return self.errors

    def reset(self):
        Asm.constants = {}
        Asm.labels = {}
        Asm.forwards = {}
        Asm.pcounter = 0
        Asm.linenum = 0
        Asm.includes = []
