# Constants:
Byte = 1
Word = 2

Immediate = 0   # Address Modes 0..4 index the mixed mode tuples
Direct = 1      # 1..2 also happen to be the operand size
Extended = 2
IndexX = 3
IndexY = 4
Inherent = 5
Relative = 6

Operators = set(['+','-','*','/','&','|','<<','>>'])   # Python driven operators
Quotes = set(['\'','\"'])
TwoOperands = set(['bset','bclr'])      # Instructions that take two operands
ThreeOperands = set(['brset','brclr'])  # Instructions that take three operands
ExtOnly = set(['asl','asr','clr','com','dec','inc','jmp',
               'lsl','lsr','neg','rol','ror','tst'])        # There is no Direct mode opcode
TwoBytes = set(['ldd','lds','ldx','ldy','addd','cpd',
                'cmpd','cpx','cmpx','cpy','cmpy','subd'])   # 16 bit Immediate instructions
Psudos = set(['org','equ','fcb','fdb','fcc','rmb','option','include']) # Psudo mnemonics for assembler operations
Branches = set([Relative, IndexX, IndexY])  # All possible modes with Relative operands

# Dictionaries of 68HC11 instruction set keyed on assembler mnemonic
# defined as a tuple of op-codes for mixed mode or as individual strings
# in the Inherent and Relative dictionaries.
# Instructions missing some of the modes (i.e. jmp, jsr, stx) use "NoOp"
# to keep the indexes into the tuple lined up.
InhCodes =  {'test':'00',
             'nop':'01',
             'idiv':'02',
             'fdiv':'03',
             'lsrd':'04',
             'lsld':'05',
             'tap':'06',
             'tpa':'07',
             'inx':'08',
             'iny':'1808',
             'dex':'09',
             'dey':'1809',
             'clv':'0A',
             'sev':'0B',
             'clc':'0C',
             'sec':'0D',
             'cli':'0E',
             'sei':'0F',
             'sba':'10',
             'cba':'11',
             'tab':'16',
             'tba':'17',
             'daa':'19',
             'aba':'1B',
             'tsx':'30',
             'tsy':'1830',
             'ins':'31',
             'pula':'32',
             'pulb':'33',
             'des':'34',
             'txs':'35',
             'tys':'1835',
             'psha':'36',
             'pshb':'37',
             'pulx':'38',
             'puly':'1838',
             'rts':'39',
             'abx':'3A',
             'aby':'183A',
             'rti':'3B',
             'pshx':'3C',
             'pshy':'183C',
             'mul':'3D',
             'wai':'3E',
             'swi':'3F',
             'nega':'40',
             'coma':'43',
             'lsra':'44',
             'rora':'46',
             'asra':'47',
             'asla':'48',
             'lsla':'48',
             'rola':'49',
             'deca':'4A',
             'inca':'4C',
             'tsta':'4D',
             'clra':'4F',
             'negb':'50',
             'comb':'53',
             'lsrb':'54',
             'rorb':'46',
             'asrb':'57',
             'aslb':'58',
             'lslb':'58',
             'rolb':'59',
             'decb':'5A',
             'incb':'5C',
             'tstb':'5D',
             'clrb':'5F',
             'xgdx':'8F',
             'xgdy':'188F',
             'stop':'CF'
             }

RelCodes = {'bra':'20',
            'brn':'21',
            'bhi':'22',
            'bls':'23',
            'bcc':'24',
            'bhs':'24',
            'bcs':'25',
            'blo':'25',
            'bne':'26',
            'beq':'27',
            'bvc':'28',
            'bvs':'29',
            'bpl':'2A',
            'bmi':'2B',
            'bge':'2C',
            'blt':'2D',
            'bgt':'2E',
            'ble':'2F',
            'bsr':'8D'
            }

# MNE(IMM,DIR,EXT,INDX,INDY)
MixModes = {'brset':('NoOp','12','NoOp','1E','181E'), 
            'brclr':('NoOp','13','NoOp','1F','181F'),
            'bset':('NoOp','14','NoOp','1C','181C'), 
            'bclr':('NoOp','15','NoOp','1D','181D'),
            'neg':('NoOp','NoOp','70','60','1860'),
            'com':('NoOp','NoOp','73','63','1863'),
            'lsr':('NoOp','NoOp','74','64','1864'),
            'ror':('NoOp','NoOp','76','66','1866'),
            'asr':('NoOp','NoOp','77','67','1867'),
            'lsl':('NoOp','NoOp','78','68','1868'),
            'asl':('NoOp','NoOp','78','68','1868'),
            'rol':('NoOp','NoOp','79','69','1869'),
            'dec':('NoOp','NoOp','7A','6A','186A'),
            'inc':('NoOp','NoOp','7C','6C','186C'),
            'tst':('NoOp','NoOp','7D','6D','186D'),
            'jmp':('NoOp','NoOp','7E','6E','186E'),
            'clr':('NoOp','NoOp','7F','6F','186F'),
            'suba':('80','90','B0','A0','18A0'),
            'subb':('C0','D0','F0','E0','18E0'),
            'cmpa':('81','91','B1','A1','18A1'),
            'cmpb':('C1','D1','F1','E1','18E1'),
            'sbca':('82','92','B2','A2','18A2'),
            'sbcb':('C2','D2','F2','E2','18E2'),
            'subd':('83','93','B3','A3','18A3'),
            'cpd':('1A83','1A93','1AB3','1AA3','CDA3'),
            'cmpd':('1A83','1A93','1AB3','1AA3','CDA3'),
            'addd':('C3','D3','F3','E3','18E3'),
            'anda':('84','94','B4','A4','18A4'),
            'andb':('C4','D4','F4','E4','18E4'),
            'bita':('85','95','B5','A5','18A5'),
            'bitb':('C5','D5','F5','E5','18E5'),
            'ldaa':('86','96','B6','A6','18A6'),
            'ldab':('C6','D6','F6','E6','18E6'),
            'staa':('NoOp','97','B7','A7','18A7'),
            'stab':('NoOp','D7','F7','E7','18E7'),
            'eora':('88','98','B8','A8','18A8'),
            'eorb':('C8','D8','F8','E8','18E8'),
            'adca':('89','99','B9','A9','18A9'),
            'adcb':('C9','D9','F9','E9','18E9'),
            'oraa':('8A','9A','BA','AA','18AA'),
            'orab':('CA','DA','FA','EA','18EA'),
            'adda':('8B','9B','BB','AB','18AB'),
            'addb':('CB','DB','FB','EB','18EB'),
            'cpx':('8C','9C','BC','AC','CDAC'),
            'cmpx':('8C','9C','BC','AC','CDAC'),
            'cpy':('188C','189C','18BC','1AAC','18AC'),
            'cmpy':('188C','189C','18BC','1AAC','18AC'),
            'jsr':('NoOp','9D','BD','AD','18AD'),
            'ldd':('CC','DC','FC','EC','18EC'),
            'std':('NoOp','DD','FD','ED','18ED'),
            'lds':('8E','9E','BE','AE','18AE'),
            'ldx':('CE','DE','FE','EE','CDEE'),
            'ldy':('18CE','18DE','18FE','1AEE','18EE'),
            'sts':('NoOp','9F','BF','AF','18AF'),
            'stx':('NoOp','DF','FF','EF','CDEF'),
            'sty':('NoOp','18DF','18FF','1AEF','18EF')
            }

kwlist = InhCodes.keys() + RelCodes.keys() + MixModes.keys() + list(Psudos)
